/*
 *   IRC - Internet Relay Chat, ircd/version.c
 *   Copyright (C) 1990 Chelsea Ashley Dyerman
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *   $
 */

/*
 * This file is generated by version.c.SH. Any changes made will go away.
 */

#include "struct.h"
#include "version.h"
#include "license.h"

char *generation = "1";
char *creation = "Fri Jul 3 2026 at 19:02:39 UTC";
#define IRCDTOTALVERSION BASE_VERSION "-" PATCH1 PATCH2 PATCH3 PATCH4 PATCH5 PATCH6 PATCH7 PATCH8 PATCH9
char *version = IRCDTOTALVERSION;
char *buildid = "6.2.6";
/* moved to s_serv.c */
char *infotext[] = 
{ 0 };

char *unrealcredits[] =
{
	"==================[ " IRCDTOTALVERSION " Credits ]===================",
	"The people on this list are people who have helped us with the",
	"development of UnrealIRCd and who have made remarkable",
	"contributions to the project.",
	" ",
	"==========================[ Donations ]==========================",
	"BlueFlame^, [Real] - ChatFIRST.com, Jameno123 - ByteHosting,",
	"Internet Services, Interlink Access Corp, Jan Knutar, ThePlayer,",
	"Headband, noriko, powerstorm.net, RedMaxima, IronHelix, xnet.org,",
	"Pierce - irc.AAcNet.org, Franky75 - Betas-Online.com, irc.vco.se,",
	"Beldock - irc.coldfront.net, Kusau - chat.tochat.org, Japsclan,",
	"WolfLord - UplinkCorp, Isaiah - irc.frogstar.us, Kedrin Milborn,",
	"Dionisios Koutsikos, Tank - irc.scifi-fans.net, irc.P2Pchat.net,",
	"Leo Zhadanovsky - irc.leozh.net, Lyote - ZodiaCIrC/DecayOnline,",
	"Jesse Lanning - Secure Technology Networks, HERZ - insiderZ.DE,",
	"James Yerge - Kronical Internet Solutions, djw - irc.perldev.org,",
	"MichaelJE2 - irc.xcelor8.com, AfterShock - irc.liquidvoltage.net,",
	"recycled-irc - www.recycled-irc.org, Alpha - irc.FireWirez.net,",
	"FrostByghte - irc.coldfront.net, SdgNem - irc.RealmOfGaming.net,",
	"MagicalTux - irc.FF.st, irc.P2Pchat.net, Ganja51 - irc.lcirc.net,",
	"^White_Magic^ - irc.chatuniverse.net, Nick - irc.plasmachat.net,",
	"Matridom - www.WinDrivers.com, anaconda - irc.lightmoon.org,", 
	"mnslinky - http://www.secure-computing.net, Justin Furnas,",
	"Andy Hansis - irc.technerd.net, Crimson - www.n00bstories.com",
	"Devin Reams, Cleggo - irc.ugcentral.net, Tillo - irc.OSirc.net,",
	"Matthew Burdine - irc.owns.us, Philip Veale - flame.tiefighter.org,",
	"Latinus - irc.lokanova.com, Vincent Guesnard - irc.rs2i.net,",
	"Windfyre IRC Network - irc.windfyre.net, SetNine - www.setnine.com,",
	"|S| - irc.chatuniverse.net, CommTech Inc. - www.commtechusa.com,",
	"CatNet IRC - irc.catslair.org, IcE-IRC - irc.ice-irc.de,",
	"Kevin Devaux, Jackal - www.fyrebird.net, Lax - insiderZ.DE,",
	"The_devil - chat.snyggast.se, Deltaanime Network - deltaanime.net,",
	"CatNet IRC network - irc.catslair.org, A-Buz.com - a-buz.com,",
	"LimonCenter Networks - www.limoncenter.com, Jean-Pierre Fournier,",
	"AloneInTheDark - irc.sleepwalkers.org, El_barto - irc.ircdit.net,",
	"Pyrexnetworks - www.pyrexnetworks.com, zunnie - www.MP-Gaming.NET,",
	"All My Data - www.amdwebhost.com, Wulfie - irc.soundsnwaves.net,",
	"eCoupons.com - www.ecoupons.com, MauritZ - irc.MindForge.org,",
	"Voodoohosting - www.voodoohosting.com,",
	"RealityBytez IRC Network - irc.realitybytez.net,",
	"LFS - www.linuxfromscratch.org, The_devil - chat.snyggast.se,",
	"Stephen - dirac.betas-online.com, NationWars - www.nationwars.com,",
	"SurrealChat.net - www.surrealchat.net, NamesDir - www.namesdir.com,",
	"Muisje - www.elitez0r.com, Jaya Sri - www.krisna.cc,",
	"DiabloClone.Org - irc.diabloclone.org, RisingNet - www.RisingNet.net,",
	"DCloneIRC.net - irc.dcloneirc.net, L.E.-Nation - irc.le-nation.de,",
	"RomeoIRC - irc.romeoirc.net, Comparison Shopping - www.order.com,",
	"Atak Trucking - www.AtakTrucking.com, www.Gamehostingprovider.com,",
	"Xzibition.com - irc.xzibition.com, www.TheBookmarkShop.com,",
	"Compare Prices - www.PriceV.com, Joe Gronlund - www.joegronlund.com,",
	"Exoware - www.exoware.net, Bitmaster - irc.synIRC.net,",
	"Angel IRC - irc.angelirc.net, Pixel - http://indavoid.net,",
	"Grandview Landscape and Masonry - www.grandviewoutdoor.com,",
	"Saibot Technologies - http://saibottechnologies.com/,",
	"IRC-IRC.de Community - irc.irc-irc.de,",
	"Mike Meisner - www.hottubcoverdepot.com, CubaChat - irc.cubachat.org,",
	"BluePromoCode - www.bluepromocode.com, DontPayFull - www.dontpayfull.com",
	"Anikwa - ukblabberbox.co.uk, xShellz - www.xshellz.com, wico,",
	"Magnolia Road Internet Cooperative, FrostWire - www.frostwire.com,",
	"Bundeld Inc. - couponlawn.com, Chameleon John - www.chameleonjohn.com,",
	"Ecigsopedia.com - www.ecigsopedia.com/halo-cigs-coupon,",
	"Hello Voucher Card - www.hellovouchercard.co.uk, Knoji - knoji.com,",
	"EveryCloud - www.everycloudtech.com, CouponRaven - couponraven.com,",
	"Dealspotr - dealspotr.com, GuitarFella.com, Coupofy - Coupofy.com,",
	"Underfloor Heating Systems - www.underfloorheatingsystems.co.uk,",
	"Project Free TV - newprojectfreetv.com, companyaccountscheck.com,",
	"Coyote Traffic - www.coyotetraffic.com, www.emailextractor14lite.com,",
	"bet365 Bonus Code - williamspromocodes.co.uk/bet365-bonus-code-sports/,",
	"Valerie Amelia Pond",
	" ",
	"==========================[ Supporters ]==========================",
	"Our support staff: Cronus, Jobe, SpaceDog, Stealth",
	" ",
	"Thanks to former supporters and anyone who has helped without",
	"officially joining the staff.",
	" ",
	"===========================[ Coders ]=============================",
	"Bram Matthys (Syzop) <syzop@unrealircd.org> - Head Coder / Project leader",
	"Krzysztof Beresztant (k4be) <k4be@unrealircd.org> - Coder",
	"Gottem <gottem@unrealircd.org> - Coder",
	"i <i@unrealircd.org> - Coder",
	" ",
	"For a full list of coders & contributors, see /INFO (/QUOTE INFO)",
	" ",
	"==================================================================",
	"Finally, we would like to thank everyone else who has helped",
	"UnrealIRCd in any way: reporting bugs, testing releases, helping",
	"others, recommending UnrealIRCd to friends, or simply just by",
	"running UnrealIRCd to ensure that we are and will continue to be",
	"the most used IRC daemon in the world.",
	"==================================================================",
	0
}; 


char unreallogo[] = 
{
32,95,32,32,32,95,32,32,32,32,32,
32,32,32,32,32,32,32,32,32,32,32,
32,32,32,32,32,32,95,32,95,95,95,
95,95,95,95,95,95,95,95,32,32,95,
95,95,95,95,32,32,32,32,32,95,32,
10,124,32,124,32,124,32,124,32,32,32,
32,32,32,32,32,32,32,32,32,32,32,
32,32,32,32,32,32,124,32,124,95,32,
32,32,95,124,32,95,95,95,32,92,47,
32,32,95,95,32,92,32,32,32,124,32,
124,10,124,32,124,32,124,32,124,95,32,
95,95,32,32,95,32,95,95,32,95,95,
95,32,32,95,95,32,95,124,32,124,32,
124,32,124,32,124,32,124,95,47,32,47,
124,32,47,32,32,92,47,32,95,95,124,
32,124,10,124,32,124,32,124,32,124,32,
39,95,32,92,124,32,39,95,95,47,32,
95,32,92,47,32,95,96,32,124,32,124,
32,124,32,124,32,124,32,32,32,32,47,
32,124,32,124,32,32,32,32,47,32,95,
96,32,124,10,124,32,124,95,124,32,124,
32,124,32,124,32,124,32,124,32,124,32,
32,95,95,47,32,40,95,124,32,124,32,
124,95,124,32,124,95,124,32,124,92,32,
92,32,124,32,92,95,95,47,92,32,40,
95,124,32,124,10,32,92,95,95,95,47,
124,95,124,32,124,95,124,95,124,32,32,
92,95,95,95,124,92,95,95,44,95,124,
95,124,92,95,95,95,47,92,95,124,32,
92,95,124,32,92,95,95,95,95,47,92,
95,95,44,95,124,10,0,85,110,114,
101,97,108,73,82,67,100,0};

char *dalinfotext[] =
    {
        " --",
        "Based on the original code written by Jarkko Oikarinen",
        "Copyright 1988, 1989, 1990, 1991 University of Oulu, Computing Center",
        "",
        "This program is free software; you can redistribute it and/or",
        "modify it under the terms of the GNU General Public License as",
        "published by the Free Software Foundation; either version 1, or",
        "(at your option) any later version.",
	"- Any name/comment should never be changed except by the one who made it -",
	"",
	"UnrealIRCd contains code developed by:",
	"Potvin       Chris Wolkowski          potvin@acestar.org",
	"RogerY       Roger Y.                 rogery@austnet.org",
	"GZ                                    gz@starchat.net",
	"binary                                ",
	"Dianora      Diane Bruce",
	"lucas", 
	"Roar Thronaas <roart@nvg.ntnu.no> added IPv6 support.",
	"",
	"",
	"The following people have helped in making the DALnet ircd",
        "that is based on irc2.8.21.mu3.2 :",
        "",
        "Russell      Russell Miller           russell@dal.net",
        "Donwulff     Jukka Santala            donwulff@dal.net",
        "Aetobatus    Michael Sawyer           aetobatus@dal.net",
        "Dalvenjah    Sven Nielsen             dalvenjah@dal.net",
        "Skandranon   Michael Graff            explorer@flame.org",
        "Barubary     -                        barubary@dal.net",
        "white_dragon Chip Norkus              wd@dal.net",
        "DuffJ        Dafydd James             duffj@dal.net",
        "taz          David Kopstain           taz@dal.net",
        "NikB         Nik Bougalis             nikb@dal.net",
        "Rakarra      -                        rakarra@dal.net",
        "DarkRot      Lucas Madar              darkrot@dal.net",
        "Studded      -                        studded@dal.net",
        "JoelKatz     David Schwartz           joelkatz@dal.net",
        "",
        "This product includes software developed by Colin Plumb.",
        "",
        "The following persons have made many changes and enhancements to the",
        "code and still know how IRC really works if you have questions about it:",
        "",
        "Run          Carlo Kid                carlo@runaway.xs4all.nl",
        "Avalon       Darren Reed              avalon@coombs.anu.edu.au",
        "msa          Markku Savela            Markku.Savela@vtt.fi",
        "Wumpus       Greg Lindahl             gl8f@virginia.edu",
        "WiZ          Jarkko Oikarinen         jto@tolsun.oulu.fi",
        "Argv         Armin Gruner Armin.Gruner@Informatik.TU-Muenchen.de",
        "",
        "Thanks to the following people for help with preparing 2.8",
        "",
        "phone        Matthew Green            phone@coombs.anu.edu.au",
        "Sodapop      Chuck Kane               ckane@ece.uiuc.edu",
        "Skygod       Matt Lyle                matt@oc.com",
        "Vesa         Vesa Ruokonen            ruokonen@lut.fi",
        "Nap          Nicolas PIOCH pioch@poly.polytechnique.fr",
        "",
        "Those who helped in prior versions and continue to be helpful:",
        "",
        "Stellan Klebom      Dan Goodwin         Mike Bolotski",
        "Ian Frechette       Markku Jarvinen     Kimmo Suominen",
        "Jeff Trim           Vijay Subramaniam   Karl Kleinpaste",
        "Bill Wisner         Tom Davis           Hugo Calendar",
        "Tom Hopkins         Stephen van den Berg",
        "Bo Adler            Michael Sandrof     Jon Solomon",
        "Jan Peterson        Helen Rose          Paul Graham",
        "",
        "Thanks also goes to those persons not mentioned here who have added",
        "their advice, opinions, and code to IRC.",
        "Thanks also to those who provide the kind sys admins who let me and",
        "others continue to develop IRC.",
        "",

        0
    };
